#!/bin/bash
#
# Build script for 'BestCrypt for Linux'
#
# This script will build 'src' and 'rpm' packages and upload it to Binary Repository
#
# Our packages are signed by http://nsk.jetico.com/redmine/attachments/470/packages_jetico.com.gpgkey key.
# (page link: http://nsk.jetico.com/redmine/projects/best-crypt-for-linux/wiki )
set -e
set -u

if [ -n "${DEBUG:-}" ] ; then
  set -x
fi

BUILD_DEPENDS_DEBIAN_6_0="debian-archive-keyring pbuilder debhelper dpkg-dev openssh-client"
BUILD_DEPENDS_DEBIAN_7_11="$BUILD_DEPENDS_DEBIAN_6_0"
BUILD_DEPENDS_UBUNTU_14_04="$BUILD_DEPENDS_DEBIAN_6_0"


### Build System Bootstrap ###
# Get 'buildsys.func.inc' from Binary Repository
SOURCES_TOP_DIR="."
BUILDSYS_SCRIPT="buildsys.func.inc"
if [ ! -s "$SOURCES_TOP_DIR/$BUILDSYS_SCRIPT" ] ; then
  # Normally, location of Binary Repository should be specified via environment,
  # see http://nsk.jetico.com/redmine/projects/bc-wipe-total-wipe-o/wiki/Build_Host_Setup
  # Default location (in builder's home directory) is for test purpose.
  BIN_REPO_DIR="${BIN_REPO_DIR:-$HOME/binary-repository}"
  BIN_REPO_SSH="${BIN_REPO_SSH:-}"
  bootstrap_get_file()
  {
    local src_file="$BIN_REPO_DIR/$1"
    local dst_file="$SOURCES_TOP_DIR/$1"
    if [ -n "${BIN_REPO_SSH:-}" ] ; then
      scp -q -o "BatchMode yes" -p "$BIN_REPO_SSH:$src_file" $dst_file
    else
      cp -a "$src_file" $dst_file
    fi
  }
  echo "Bootstrap: getting '$BUILDSYS_SCRIPT' from $BIN_REPO_SSH $BIN_REPO_DIR"
  bootstrap_get_file "$BUILDSYS_SCRIPT"
fi
. $SOURCES_TOP_DIR/$BUILDSYS_SCRIPT
### end of bootstrap ###



### Configuration ###
ARCH=`uname -m`

BUILD_DIR="build_deb_tmp"

# use new layout of Binary Repository (with components/ subdirectory)
BIN_REPO_COMPONENTS_SUBDIR="components"




### Pre-build checks ###

# check binary repository availability
ping_repository
check_build_depends
set_source_control_variables



### Constants ###

# Define SSH access for a remote host with APT repository.
#
# Requires: 'reprepo' been installed on the remote host.
#           (used by 'contrib/put_package_in_repo.sh' script)
#APT_REPO_SSH="binrepo@file"
#APT_REPO_DIR="/home/binrepo/binary-repository/linux-repository"
APT_REPO_SSH="uploader@umbrella.jetico"
APT_REPO_DIR="/srv/Public/packages_dev/${SOURCE_CONTROL_PROJ}_${SOURCE_CONTROL_BRANCH}"


### Get components depending on ###


# Download source package

if [ "${TEST_ONLY:-}" = "YES" ] ; then
  # TESTONLY

  # fixed source directory, easy to investigate
  TMP_SOURCES="/tmp/src_build_binaries_test"

  # create only once
  if [ ! -d $TMP_SOURCES ] ; then
    mkdir $TMP_SOURCES
    ( cd $TMP_SOURCES ; apt-get source --download-only hello )
  fi

else

  # Normal mode
TMP_SOURCES=`mktemp -d /tmp/src_bestcrypt_src_XXXXXX`

download_component_to_directory "src" "$TMP_SOURCES" "all"

fi




### Build ###


# Put package installation instruction as index.html
INSTALL_HTML="index.html"
cat << EOF > "$INSTALL_HTML"
<html>
 <head>
  <title>BestCrypt installation instruction</title>
 </head>
 <body>
  <pre>
  `cat INSTALL.packages`
  </pre>
 </body>
</html>
EOF
echo "===> Refreshing index.html at APT_REPO ($APT_REPO_SSH:$APT_REPO_DIR/)"
ssh $APT_REPO_SSH "mkdir -p $APT_REPO_DIR"
scp "$INSTALL_HTML" "$APT_REPO_SSH:$APT_REPO_DIR/"

# Script for server-side adding package into linux-repository
scp contrib/put_package_in_repo.sh "$APT_REPO_SSH:"


# Extract debian source package, fix version and build for specified distribution
#  Args:
#  $1 - distro name (ubuntu/precise, debian/squeeze)
#  $2 - directory with *.dsc file to rebuild
build_source_debian_package_with_pbuilder()
{
  local dist="$1"
  local workdir="$2"
  local distribution=`echo $1 | cut -d/ -f1`
  local dist_codename=`echo $1 | cut -d/ -f2`

  echo
  echo
  echo
  echo "===> $dist: Extract debian source package..."
  echo
  echo

  echo -e " Filelist of workdir $workdir:"
  ls -l $workdir

  # get absolute path to .dsc file
  DSC=""
  for i in `find $workdir -maxdepth 1 -name '*.dsc'` ; do
    if [ -z "$DSC" ] ; then
      DSC="$i"
    else
      echo "Error: must be only one .dsc file (found second: $i)"
      exit 1
    fi
  done

  local distname_dir="$workdir/$dist_codename"

  # already built
  if [ -d "$distname_dir" ] ; then
    rm -rf "$distname_dir"
    # sudo rm -rf "$distname_dir"
  fi

  mkdir "$distname_dir"
  (
    cd $distname_dir
    dpkg-source -x "$DSC"
  )

  # get absolute path of extracted <package>-<version> directory
  local debsrcextracted_dir=`find "$distname_dir" -maxdepth 1 -type d -name '*-*'`


  echo "===> $dist: Build debian binary package..."
  (
    cd $debsrcextracted_dir


    echo -e "\n\n  We are in debsrcextracted_dir $debsrcextracted_dir now, filelist:"
    ls -l


    # will fail to pack it back???
    #DIR=`find -type d -maxdepth 1 -name '*-*'`
    #test -n "$DIR" || { echo "Error: $0: unable to get exctracted debian package directory" ; exit 1 ; }
    #cd $DIR
    # previous method: just build on host system: dpkg-buildpackage -B

    # this was tested on Debian 6.0
    # with following packages from 'sid' (jul 2014):
    #   * ubuntu-archive-keyring_2012.05.19-1_all.deb
    #   * debootstrap_1.0.60_all.deb

    # add $dist_codename to package version to avoid .deb files clash in APT repository (pool/ folder)
    DEB_VER=`dpkg-parsechangelog -ldebian/changelog -o0 -c1 | awk '/^Version:/ { print $2 }'`

    # add distribution codename to .deb version
    sed -i "s/${DEB_VER}/${DEB_VER}.${dist_codename}/" debian/changelog



    # Create pbuilder environment if it is not created for this dist yet
    BASETGZ="/var/cache/pbuilder/$dist_codename.tgz"

    echo "Building with pbuilder using base tgz: $BASETGZ"

    if [ ! -s "$BASETGZ" ] ; then # if previous 'pbuilder create' failed this file will exist but empty

      echo "Base tgz is absent, rebuild it"

      if [ ! -s /usr/share/debootstrap/scripts/$dist_codename ] ; then

cat <<EOF

Error: $0: seems you are using an old version of debootstrap

Hint: You should install the lates possible from 'sid' repository:

   https://packages.debian.org/debootstrap

EOF
        rm -rf $debsrcextracted_dir
        exit 1
      fi

      # hint: keep distribution codenames in sync with contrib/put_package_in_repo.sh
      case $distribution in
        debian)
          MIRROR="http://mirror.yandex.ru/debian/"
          KEYRING="/usr/share/keyrings/debian-archive-keyring.gpg"
          ;;
        ubuntu)
          MIRROR="http://ru.archive.ubuntu.com/ubuntu/"
          KEYRING="/usr/share/keyrings/ubuntu-archive-keyring.gpg"
          ;;
        oldubuntu)
          MIRROR="http://old-releases.ubuntu.com/ubuntu/"
          KEYRING="/usr/share/keyrings/ubuntu-archive-keyring.gpg"
          ;;
        *)
          echo "Error: $0: please add MIRROR= definition to this script for dist '$dist'"
          exit 1
      esac

      if [ ! -s $KEYRING ] ; then
        cat <<EOF

Error: $0: there is no keyring file $KEYRING.

You should install appropriate debian-archive-keyring or
ubuntu-archive-keyring package.

On Debian you can get it directly from 'sid': https://packages.debian.org/ubuntu-archive-keyring

EOF
        rm -rf $debsrcextracted_dir
        exit 1
      fi

      sudo pbuilder create \
        --distribution $dist_codename \
        --mirror $MIRROR \
        --debootstrapopts --keyring=$KEYRING \
        --basetgz "$BASETGZ"
    else
      echo "Base tgz exists, rebuild using it:"
      ls -l "$BASETGZ"
    fi


    echo
    echo
    echo -e "\n\n Trying to build with command: pdebuild -- --basetgz $BASETGZ --buildresult $distname_dir"
    echo "   in directory `pwd`"
    echo -e "  Result should be put to $distname_dir \n\n"

    if ! sudo pdebuild -- --basetgz "$BASETGZ" --buildresult "$distname_dir" ; then
      # may be we need to do 'pbuilder update' to refresh the packages list.
      # we should do it anyway at least once per week.
    if find "$BASETGZ" -mtime +7 >/dev/null ; then
      echo
        echo "Info: $BASETGZ is older than 7 days. Try to update the system image and re-try build:"
        date
        ls -l "$BASETGZ"
      echo
      sudo pbuilder update --basetgz "$BASETGZ"

        # and try to re-build
        sudo pdebuild -- --basetgz "$BASETGZ" --buildresult "$distname_dir"
      fi
    fi

    # after sudo *.debs will belong to 'root'
    sudo find $distname_dir -maxdepth 1  -type f -exec chown `whoami` {} \;

    echo -e "\n\n\n   Build process is finished."
    echo "Content of workdir $workdir:"
    ls -l $workdir

    echo -e "\nContent of debsrcextracted_dir $debsrcextracted_dir:"
    ls -l $debsrcextracted_dir

    echo -e "\nContent of distname_dir $distname_dir:"
    ls -l $distname_dir

    rm -rf $debsrcextracted_dir
    echo -e "  unpackdir was cleaned-up (removed completely)\n\n"
  )
}





# please don't use it. Use APT repository instead
upload_deb_to_binary_repository()
{
  ### Upload to Binary Repository (deprecated, turned off since Apr 2015) ###
  echo "===> $dist: Upload to Binary Repository"
  TMP_COMP=`mktemp -d /tmp/bestcrypt_deb_XXXXXX`

  $CP $DEB_PKG $TMP_COMP/

  echo "$SOURCE_CONTROL_REV" > $TMP_COMP/revision.txt
  echo -e "Recent changes:\n\n${SOURCE_CONTROL_LOG}" > $TMP_COMP/recent_changes.txt

  # build host info
  write_buildinfo_file "$TMP_COMP/buildinfo.txt"

  FILELIST="$DEB_PKG_FN revision.txt recent_changes.txt buildinfo.txt"

  # create digest file
  ( cd $TMP_COMP ; md5sum $FILELIST > md5sums.txt )

  # TODO: do we need this binaries in our Binary Repository?
  # It does a little sense because it is better to install packages with APT.

  # no, we do not need it (apr 2015). No .deb anymore in Binary Repository (they are not a components either)
  # upload_component_from_directory "deb_${dist_codename}" $TMP_COMP
}


# Args:
#  $1 - target distribution (ubuntu/precise, debian/squeeze)
#  $2 - distribution directory where .debs for this distribution are resided
upload_debs_to_apt_repository()
{
  local dist="$1"
  local distname_dir="$2"
  local distribution=`echo $1 | cut -d/ -f1`
  local dist_codename=`echo $1 | cut -d/ -f2`

  # Upload binary packages into APT repository
  echo "===> $dist: Upload to APT repository"

  echo -e "\n\n\n  Directory with packages: $distname_dir  \n\n\n"
  echo "   file listing:"
  ls -l $distname_dir

  # Find Debian binary packages
  local count=0
  local DEB_PKG_FN=""

  for a in `find $distname_dir/ -maxdepth 1 -name '*.deb'` ; do
    DEB_PKG_FN="$DEB_PKG_FN `basename $a`"
    count=$(( $count + 1 ))
  done

  echo -e "  Found packages are:  $DEB_PKG_FN    \n\n\n"

  if [ "$count" -lt "1" ] ; then
    echo "Error: no .deb packages found after build"
    exit 1
  fi

  scp contrib/put_package_in_repo.sh $APT_REPO_SSH:

  # Upload to remove APT repository
  local TMP_INCOMING=`ssh $APT_REPO_SSH  mktemp --directory /tmp/put_package_incoming_XXXXXX`    # create temp dir on remote host

  ( cd $distname_dir ; scp ${DEB_PKG_FN} ${APT_REPO_SSH}:${TMP_INCOMING}/ )
  local REMOTE_HOME=`ssh $APT_REPO_SSH  pwd`
  ssh $APT_REPO_SSH "cd $TMP_INCOMING ; DEBUG=${DEBUG:-} ${REMOTE_HOME}/put_package_in_repo.sh -d $distribution -c $dist_codename -b $APT_REPO_DIR ${DEB_PKG_FN}"
  res="$?"

  # clean up
  # ssh $APT_REPO_SSH "rm -rf $TMP_INCOMING"

  if [ "$res" != "0" ] ; then
    echo "Warning: put_package_in_repo.sh finished unsuccessfully: exit code $?"
    exit 1
  fi

  echo "Packages ${DEB_PKG_FN} was uploaded to  $APT_REPO_DIR"
}


# https://wiki.ubuntu.com/Releases
# https://www.debian.org/releases/
#
# Listed (and built) in order:
# first line: most supported the latest stable/LTS versions
# then from more recent to older
DISTS="ubuntu/trusty ubuntu/xenial debian/jessie
       ubuntu/xenial     debian/sid
       ubuntu/wily       debian/stretch
       ubuntu/vivid      debian/wheezy
       ubuntu/precise
      "

# step 1: build .deb packages
for dist in $DISTS
do

  echo -e "\n\n\n  build_source_debian_package_with_pbuilder $dist $TMP_SOURCES   \n\n\n"
  build_source_debian_package_with_pbuilder "$dist" "$TMP_SOURCES"
done


# step 2: upload to APT repo
for dist in $DISTS
do
  dist_codename=`echo $dist | cut -d/ -f2`

  echo -e "\n\n\n  upload_debs_to_apt_repository $dist $TMP_SOURCES/$dist_codename  \n\n\n"
  upload_debs_to_apt_repository "$dist" "$TMP_SOURCES/$dist_codename"
done


# clean-up
test -z "${TEST_ONLY:-}" && rm -rf "$TMP_SOURCES/"
