#include <string.h>
#include "cSymAlg.h"
#include "../endian.h"

cSymAlg::cSymAlg()
{
	m_iv = NULL;
	m_prev = NULL;
	m_iv_offset = 0;

	setBCDriverCompatible( eOriginial );
}


cSymAlg::~cSymAlg()
{
	if (m_iv != NULL) safeFree( m_iv );
	m_iv = NULL;
	
	if (m_prev != NULL) safeFree( m_prev );
	m_prev = NULL;

	m_iv_offset = 0;
}

// Encryption and decription in Cipher Feedback mode

int cSymAlg::initCFB( const BYTE *iv ) // iv - Initial vector
{
	int blockSize, keySize;
	
//	if ( iv == NULL ) return sae_invalidBlockSize;

	if (!getAlgConstants( &blockSize, &keySize )) return sae_internalError;

	if (!m_iv) m_iv = safeAlloc( blockSize );
	if (!m_iv) return sae_memoryAllocationError; // sae_internalError;

	if (!m_prev) m_prev = safeAlloc( blockSize );
	if (!m_prev)
	{
		safeFree( m_iv );
		return sae_memoryAllocationError; // sae_internalError;
	}

	if ( iv == NULL )
		memset( m_iv, 0, blockSize );
	else
		memcpy( m_iv, iv, blockSize );

	m_iv_offset = 0;

	return sae_ok;
}



int cSymAlg::encryptCFB(
		const BYTE *inBf,		// Input plaintext buffer. 
		BYTE *outBf,	// Output encrypted buffer
		int  bfSize		// In/out buffers' size.
		)
{
	int blockSize, keySize, i, iRes;
		
	if (!getAlgConstants( &blockSize, &keySize )) return sae_internalError;

	//
	//	v. 1.04.8.3 - 20.10.2016
	//	Checks block size here, otherwise following 'm_iv_offset %= blockSize;' crushes the app.
	//
	if ( blockSize == 0 )
	{
		//	Dummy algorithm has blockSize == 0
		//		return sae_invalidBlockSize;
		memcpy( outBf, inBf, bfSize );
		return sae_ok;
	}

	for (i=0; i<bfSize; i++, m_iv_offset++ )
	{
		m_iv_offset %= blockSize;

		if (m_iv_offset == 0)
		{
			memcpy( m_prev, m_iv, blockSize );
			iRes = encryptECB( m_iv, m_iv, blockSize );
			if (iRes != sae_ok) return iRes;
		}

		outBf[i] = inBf[i] ^ m_iv[m_iv_offset];
		m_iv[m_iv_offset] = outBf[i];
	}

	return sae_ok;
}


int cSymAlg::decryptCFB(
		const BYTE *inBf,	// Input plaintext buffer. 
		BYTE *outBf,		// Output encrypted buffer
		int  bfSize			// In/out buffers' size.
		)
{
	int  blockSize, keySize, i, iRes;
	BYTE bTmp;
		
	if (!getAlgConstants( &blockSize, &keySize ))
		return sae_internalError;

	//
	//	v. 1.04.8.3 - 20.10.2016
	//	Checks block size here, otherwise following 'm_iv_offset %= blockSize;' crushes the app.
	//
	if ( blockSize == 0 )
	{
		//	Dummy algorithm has blockSize == 0
		//		return sae_invalidBlockSize;
		memcpy( outBf, inBf, bfSize );
		return sae_ok;
	}

	for (i=0; i<bfSize; i++, m_iv_offset++ )
	{
		m_iv_offset %= blockSize;

		if (m_iv_offset == 0)
		{
			memcpy( m_prev, m_iv, blockSize );
			iRes = encryptECB( m_iv, m_iv, blockSize );
			if (iRes != sae_ok) return iRes;
		}

		bTmp = inBf[i]; // We need bTmp because inBf may be the same as outBf
		outBf[i] = m_iv[m_iv_offset] ^ inBf[i];
		m_iv[m_iv_offset] = bTmp;
	}

	return sae_ok;
}
/*
	static void
pgpCFBSync(CFBInterleaveStruct *cfb)
{
	PGPSize	blockSize;
	PGPSize	bufLeft		= cfb->bufLeft;

	(void)PGPGetSymmetricCipherSizes( cfb->symmetricRef, NULL, &blockSize );
	
	if (bufLeft)
	{
		pgpCopyMemory( cfb->iv, cfb->iv + bufLeft, blockSize - bufLeft);
		pgpCopyMemory( cfb->prev + blockSize - bufLeft, cfb->iv, bufLeft);
		cfb->bufLeft	= 0;
	}
}
*/

int cSymAlg::syncCFB()
{
	int  blockSize, keySize, bytesLeft;
		
	if (!getAlgConstants( &blockSize, &keySize )) return sae_internalError;

	bytesLeft = blockSize - m_iv_offset;

	if ( bytesLeft )
	{
		memmove( m_iv, &(m_iv[ bytesLeft ]), m_iv_offset );
		memcpy( &(m_prev[m_iv_offset]), m_iv, bytesLeft );
		m_iv_offset = blockSize;
	}

	return sae_ok;
}

static inline void memxor( BYTE *r, const BYTE *v0, const BYTE *v1, int size )
{
	//
	// Compilator will optimize this even for Solaris, see note below.
	// 
	int i = 0;
	for( i = 0; i + 7 < size; i += 8 )
	{
		r[i+0] = v0[i+0] ^ v1[i+0];
		r[i+1] = v0[i+1] ^ v1[i+1];
		r[i+2] = v0[i+2] ^ v1[i+2];
		r[i+3] = v0[i+3] ^ v1[i+3];
		r[i+4] = v0[i+4] ^ v1[i+4];
		r[i+5] = v0[i+5] ^ v1[i+5];
		r[i+6] = v0[i+6] ^ v1[i+6];
		r[i+7] = v0[i+7] ^ v1[i+7];
	}

	for( ; i < size; i++ )
		r[i] = v0[i] ^ v1[i];
}

class cMemBlock
{
public:

	BYTE *m_ptr;
	int m_size;

	cMemBlock( int blockSize )
	{
		m_ptr = (BYTE*)SMalloc( blockSize );
		m_size = ( m_ptr == 0 ? 0 : blockSize );
	}
	
	~cMemBlock()
	{
		if ( m_ptr != NULL )
			SFree( m_ptr );
	}
};


/*  Macroses to transform array of 4 bytes to 32-bit dwords
 *  (and reverse) without depending on the Little-Endian or
 *  Big-Endian processor's architecture
 */

#define BYTES_TO_DWORD(b,d) (d =((DWORD)(*((b)++))), \
                             d|=((DWORD)(*((b)++)))<< 8, \
                             d|=((DWORD)(*((b)++)))<<16, \
                             d|=((DWORD)(*((b)++)))<<24)

#define DWORD_TO_4BYTES(d,b) (*((b)++)=(BYTE)(((d)    )&0xff), \
                              *((b)++)=(BYTE)(((d)>> 8)&0xff), \
                              *((b)++)=(BYTE)(((d)>>16)&0xff), \
                              *((b)++)=(BYTE)(((d)>>24)&0xff))

// The macroses above were used in driver versions, they make sens for little-endian architecture only
// So will just change byte order for LE and skip this operation in BE environment.

#ifdef _LITTLE_ENDIAN_DEFINED
	#define changeByteOrder_LE( src, dst, len ) changeByteOrder( src, dst, len )
	#define swapBytes_LE( bf, len ) swapBytes( bf, len )
#elif defined(_BIG_ENDIAN_DEFINED)
	#define changeByteOrder_LE( src, dst, len ) ( ( dst != src ) ?  memcpy( dst, src, len ) : 0 )
	#define swapBytes_LE( bf, len ) (void)0;
#else
	#error "machine is undefined endian"
#endif

static inline void swapBytes( BYTE &b1, BYTE &b2 )
{
	b1 ^= b2;
	b2 ^= b1;
	b1 ^= b2;
}

static inline void swapBytes( BYTE *bf, size_t len )
{
	for( size_t i = 0; i < len; i+= 4 )
	{
		swapBytes( bf[i+0], bf[i+3] );
		swapBytes( bf[i+1], bf[i+2] );
	}
}

static inline void changeByteOrder( const BYTE *src, BYTE *dst, size_t len )
{
	size_t i = 0;

	bool bSame = ( src == dst );

	if ( bSame )
		swapBytes( dst, len );
	else
	{
		for( ; i < len; i+= 4 )
		{
			dst[i+0] = src[i+3];
			dst[i+1] = src[i+2];
			dst[i+2] = src[i+1];
			dst[i+3] = src[i+0];
		}
	}
}

int cSymAlg::encryptCBC_BCDriverCompatible( const BYTE *sync, const BYTE *inBf, BYTE *outBf, int bfSize )
{
	int blockSize, keySize, i;
	
	if (!getAlgConstants( &blockSize, &keySize ))
		return sae_internalError;

	if (!blockSize)
		return sae_invalidBlockSize;

	if ((bfSize % blockSize) != 0)
		return sae_invalidBlockSize;

	cMemBlock mb( blockSize );

	if ( mb.m_ptr == NULL )
		return sae_memoryAllocationError;

	BYTE *internalBuf = mb.m_ptr;

	memcpy( internalBuf, sync, blockSize );

	for ( i=0; i < bfSize; i += blockSize )
	{
		memxor( internalBuf, internalBuf, inBf + i, blockSize );
		swapBytes_LE( internalBuf, blockSize );
		encryptECB( internalBuf, outBf + i, blockSize );
		swapBytes_LE( outBf + i, blockSize );
		memcpy( internalBuf, outBf + i, blockSize );
	}

	// Ignore checking of decrypt result in loops, morover all algorithm valies and boundaries are checked above.
	return sae_ok;
}

int cSymAlg::encryptCBC(
	const BYTE *sync,     // Initial vector
	const BYTE *inBf,		// Input plaintext buffer. 
	BYTE *outBf,	// Output encrypted buffer
	int  bfSize		// In/out buffers' size.
	)
{
	int blockSize, keySize, i;
	
	if (!getAlgConstants( &blockSize, &keySize ))
		return sae_internalError;

	if (!blockSize)
		return sae_invalidBlockSize;

	if ((bfSize % blockSize) != 0)
		return sae_invalidBlockSize;

	cMemBlock mb( blockSize * 2);

	if ( mb.m_ptr == NULL )
		return sae_memoryAllocationError;

	BYTE *internalBuf = mb.m_ptr;
	bool bInvalidPtr = false;
	BYTE *encOut = mb.m_ptr + blockSize;

#ifdef _ANDROID
	if ( (int)outBf & 0x7) // Invalid alignment for android leads to crash
		bInvalidPtr = true;
#endif

	memcpy( internalBuf, sync, blockSize );

	if ( bInvalidPtr )
	{
		for ( i=0; i < bfSize; i += blockSize )
		{
			memxor( internalBuf, internalBuf, inBf + i, blockSize );
			memcpy( encOut, outBf + i, blockSize );
			encryptECB( internalBuf, encOut, blockSize );
			memcpy( internalBuf, encOut, blockSize );
			memcpy( outBf + i, encOut, blockSize );
		}
	}
	else
	{
		for ( i=0; i < bfSize; i += blockSize )
		{
			memxor( internalBuf, internalBuf, inBf + i, blockSize );
			encryptECB( internalBuf, outBf + i, blockSize );
			memcpy( internalBuf, outBf + i, blockSize );
		}
	}

	// Ignore checking of decrypt result in loops, morover all algorithm valies and boundaries are checked above.
	return sae_ok;
}

int cSymAlg::decryptCBC_BCDriverCompatible(
		const BYTE *sync,	// Initial vector
		const BYTE *inBf,	// Input encrypted buffer.
		BYTE *outBf,		// Output plaintext buffer.
		int  bfSize			// In/out buffers' size.
	)
{
	int blockSize, keySize, i;
	
	if (!getAlgConstants( &blockSize, &keySize ))
		return sae_internalError;

	if (!blockSize)
		return sae_invalidBlockSize;

	if ((bfSize % blockSize) != 0)
		return sae_invalidBlockSize;

	cMemBlock mb( blockSize * 2 );
	if ( mb.m_ptr == NULL )
		return sae_memoryAllocationError;

	BYTE *oldBuf	= mb.m_ptr;
	BYTE *ivector	= mb.m_ptr + blockSize;

	memcpy( ivector, sync, blockSize );

	for ( i=0; i < bfSize; i += blockSize )
	{
		memcpy( oldBuf, inBf + i, blockSize );

		changeByteOrder_LE( inBf + i, outBf + i, blockSize );
		decryptECB( outBf + i, outBf + i, blockSize );
		swapBytes_LE( outBf + i, blockSize );

		memxor( outBf + i, outBf + i, ivector, blockSize );
		memcpy( ivector, oldBuf, blockSize ); 
	}

	// Ignore checking of decrypt result in loops, morover all algorithm valies and boundaries are checked above.
	return sae_ok;
}

//
// Note, The function is modified 7 Oct 2008.
// Solaris code cannot properly copy UINT buffer if it's boundary is not divided by 4
// So the places have been changed to memcpy and memxor here.
//
int cSymAlg::decryptCBC(
		const BYTE *sync,	// Initial vector
		const BYTE *inBf,	// Input encrypted buffer.
		BYTE *outBf,		// Output plaintext buffer.
		int  bfSize			// In/out buffers' size.
	)
{
	int blockSize, keySize, i;
	
	if (!getAlgConstants( &blockSize, &keySize ))
		return sae_internalError;

	if (!blockSize)
		return sae_invalidBlockSize;

	if ((bfSize % blockSize) != 0)
		return sae_invalidBlockSize;

	cMemBlock mb( blockSize * 2 );
	if ( mb.m_ptr == NULL )
		return sae_memoryAllocationError;

	BYTE *oldBuf	= mb.m_ptr;
	BYTE *ivector	= mb.m_ptr + blockSize;

	memcpy( ivector, sync, blockSize );

	bool bSameBuffers = ( inBf == outBf );

	if ( bSameBuffers )
	{
		for ( i=0; i < bfSize; i += blockSize )
		{
			memcpy( oldBuf, inBf + i, blockSize );
			decryptECB( inBf + i, outBf + i, blockSize );
			memxor( outBf + i, outBf + i, ivector, blockSize );
			memcpy( ivector, oldBuf, blockSize ); 
		}
	}
	else
	{
		for ( i=0; i < bfSize; i += blockSize )
		{
			decryptECB( inBf + i, outBf + i, blockSize );
			memxor( outBf + i, outBf + i, ivector, blockSize );
			memcpy( ivector, inBf + i, blockSize ); 
		}
	}

	// Ignore checking of decrypt result in loops, morover all algorithm valies and boundaries are checked above.
	return sae_ok;
}

int cSymAlg::decryptCFB	(
		const BYTE *sync,	// Initial vector
		const BYTE *inBf,	// Input encrypted buffer.
		BYTE *outBf,		// Output plaintext buffer.
		int  bfSize			// In/out buffers' size.
	)
{
	initCFB( sync );

	return decryptCFB( inBf, outBf, bfSize );
}

int cSymAlg::encryptCFB(
		const BYTE *sync,	// Initial vector
		const BYTE *inBf,	// Input encrypted buffer.
		BYTE *outBf,		// Output plaintext buffer.
		int  bfSize			// In/out buffers' size.
	)
{
	initCFB( sync );

	return encryptCFB( inBf, outBf, bfSize );
}
