/*
 * Adapted from hmac.c (HMAC-MD5) for use by SHA1.
 * by <mcr@sandelman.ottawa.on.ca>. Test cases from RFC2202.
 * source -> http://www.opensource.apple.com/source/freeradius/freeradius-11/freeradius/src/lib/hmacsha1.c
 *
 */

/*
** Function: hmac_sha1
*/


#include <string.h>
#include <sys/types.h>
#include "../hashAlgs/SHA1/Sha1.h"
#include "hmac_funcs.h"
#ifdef _usePTB
#include "../SMemory/SBlocks.h"
#endif

/*
function hmac (key, message)
    if (length(key) > blocksize) then
        key = hash(key) // keys longer than blocksize are shortened
    end if
    if (length(key) < blocksize) then
        key = key || [0x00 * (blocksize - length(key))] // keys shorter than blocksize are zero-padded (where || is concatenation)
    end if
   
    o_key_pad = [0x5c * blocksize] ^ key // Where blocksize is that of the underlying hash function
    i_key_pad = [0x36 * blocksize] ^ key // Where ^ is exclusive or (XOR)
   
    return hash(o_key_pad || hash(i_key_pad || message)) // Where || is concatenation
end function
*/

/*
unsigned char*  text;                pointer to data stream
int             text_len;            length of data stream
unsigned char*  key;                 pointer to authentication key
int             key_len;             length of authentication key
unsigned char*  digest;              caller digest to be filled in
*/

int hmac_sha1(
		const unsigned char *text, int text_len,
		const unsigned char *key , int key_len,
		unsigned char *digest
	)
{

	unsigned char k_ipad[65];    /* inner padding - key XORd with ipad  */
	unsigned char k_opad[65];    /* outer padding - key XORd with opad  */
	unsigned char tk[20];
	int i;

	/* if key is longer than 64 bytes reset it to key=SHA1(key) */
	if (key_len > 64) 
	{
		SHA1MakeDigest( key, key_len, tk );

		key = tk;
		key_len = 20;
	}

	/* start out by storing key in pads */
	memset( k_ipad, 0, sizeof(k_ipad));
	memset( k_opad, 0, sizeof(k_opad));
	memcpy( k_ipad, key, key_len);
	memcpy( k_opad, key, key_len);

	/* XOR key with ipad and opad values */
	for (i = 0; i < 64; i++)
	{
		k_ipad[i] ^= 0x36;
		k_opad[i] ^= 0x5c;
	}

#ifdef _usePTB

	cPTB ptb;
	ptb.WriteBytes( k_ipad, 64 );
	ptb.WriteBytes( text, text_len );
	SHA1MakeDigest( ptb.buffer, ptb.bufferSize, digest );

	ptb.WriteBytes( k_opad, 64, true );
	ptb.WriteBytes( digest, 20 );
	SHA1MakeDigest( ptb.buffer, ptb.bufferSize, digest );
#else

	void *ctx = NULL;
	SHA1Allocate( &ctx );
	if ( ctx == NULL )
		return -1;

	SHA1Init(ctx);                   /* init context for 1st
										  * pass */
	SHA1Process(ctx, k_ipad, 64);      /* start with inner pad */
	SHA1Process(ctx, text, text_len); /* then text of datagram */
	SHA1Final(ctx, digest);          /* finish up 1st pass */
	SHA1Init(ctx);                   /* init context for 2nd
										  * pass */
	SHA1Process(ctx, k_opad, 64);     /* start with outer pad */
	SHA1Process(ctx, digest, 20);     /* then results of 1st
										  * hash */
	SHA1Final(ctx, digest);          /* finish up 2nd pass */

	SHA1Free( &ctx );
#endif

	return 0;
}

/*
  key =         0xAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA

  key_len       16 bytes
  data =        0xDDDDDDDDDDDDDDDDDDDD...
                ..DDDDDDDDDDDDDDDDDDDD...
                ..DDDDDDDDDDDDDDDDDDDD...
                ..DDDDDDDDDDDDDDDDDDDD...
                ..DDDDDDDDDDDDDDDDDDDD
  data_len =    50 bytes
  digest =      0x56be34521d144c88dbb8c733f0e8b3f6

TEST CASE 4
        Key Length = 20
        KEY =
                0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa 0xaa
                0xaa 0xaa 0xaa 0xaa
        Data Length = 50
        DATA =
                0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd
                0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd
                0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd 0xdd
                0xdd 0xdd
        Digest (lsb first, msb last) =
                0x12 0x5d 0x73 0x42 0xb9 0xac 0x11 0xcd 0x91 0xa3
                0x9a 0xf4 0x8a 0xa1 0x7b 0x4f 0x63 0xf1 0x75 0xd3
*/

int hmac_sha1_autotest()
{
	unsigned char key[20];
	memset( key, 0xAA, sizeof( key ) );
	int key_len = sizeof( key );

	unsigned char text[50];
	memset( text, 0xDD, sizeof( text ) );
	int text_len = sizeof( text );

	unsigned char digest[20];

	int res = hmac_sha1(text, text_len, key, key_len, digest);
	if ( res != 0 )
		return res;

	unsigned char bfRes[20] = {
        0x12, 0x5d, 0x73, 0x42, 0xb9, 0xac, 0x11, 0xcd, 0x91, 0xa3,
        0x9a, 0xf4, 0x8a, 0xa1, 0x7b, 0x4f, 0x63, 0xf1, 0x75, 0xd3,
	};

	res = memcmp( bfRes, digest, sizeof( digest ) );

	return res;
}
