#include "pkcs12_structs.h"
#include "pkcs12_funcs.h"
#include "../hashAlgs/hashAlgs.h"
#include "../PGPPkts/pktdefs.h"
#include "hmac_funcs.h"
#include "oidFuncs.h"

int sPbeParam::prepareForWritting( const char *pbeAlgOID, unsigned int saltSize, unsigned int iterations )
{
	return x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			oid.prepareForWritting( pbeAlgOID ) +
			seqParams.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				octSalt.prepareForWritting ( asn1tag_OCTET_STRING, saltSize ) +
				intIterations.prepareForWritting( iterations )
				)
		);
}

int sPbeParam::writeHeader( BYTE *buff, int buffSize, int *pPos  )
{
	int hdrSize = x509_struct::writeHeader( buff, buffSize, pPos );
	if ( hdrSize == 0 ) return 0;

	int pos = 0;
	if ( oid.writeHeader( B, L, &pos ) == 0 ) return 0;
	oid.writeBody();
	if ( seqParams.writeHeader( B, L, &pos ) == 0 ) return 0;
	if ( pos != (int)L ) return 0;

	pos = 0;
	if ( octSalt.writeHeader( seqParams.B, seqParams.L, &pos ) == 0 ) return 0;
	if ( intIterations.writeHeader( seqParams.B, seqParams.L, &pos ) == 0 ) return 0;
	intIterations.writeBody();
	if ( pos != (int)seqParams.L ) return 0;

	return hdrSize;
}

int sPbeParam::writeBody( const BYTE *salt, unsigned int saltSize )
{
	if ( octSalt.writeBody( salt, saltSize ) != x509e_ok ) return x509e_internalError;
	return x509e_ok;
}


int sEncryptedContentInfo::prepareForWritting( const char *pbeAlgOID, unsigned int saltSize, unsigned int iterations, unsigned int bodySize )
{
	return
		x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, 
			version.prepareForWritting( 0 ) +
			seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				oid.prepareForWritting( szP7data ) +
				pbe.prepareForWritting( pbeAlgOID, saltSize, iterations ) +
				content.prepareForWritting( asn1tag_OPTIONAL, bodySize )
				)
			);
}

int sEncryptedContentInfo::writeHeader( BYTE *buff, int buffSize, int *pPos )
{
	int hdrSize = x509_struct::writeHeader( buff, buffSize, pPos );
	if ( hdrSize == 0 ) return 0;

	int pos = 0;
	if ( version.writeHeader( B, L, &pos ) == 0 ) return 0;
	version.writeBody();
	if ( seq.writeHeader( B, L, &pos ) == 0 ) return 0;
	if ( pos != (int)L ) return 0;

	pos = 0;
	if ( oid.writeHeader( seq.B, seq.L, &pos ) == 0 ) return 0;
	oid.writeBody();
	if ( pbe.writeHeader( seq.B, seq.L, &pos ) == 0 ) return 0;
	if ( content.writeHeader( seq.B, seq.L, &pos ) == 0 ) return 0;
	if ( pos != (int)seq.L ) return 0;

	return hdrSize;
}

int sEncryptedContentInfo::writeBody( const BYTE *salt, unsigned int saltSize, const BYTE *body, unsigned int bodySize )
{
	if ( pbe.writeBody( salt, saltSize ) != x509e_ok ) return x509e_internalError;

	if ( ( body != NULL ) &&
		 ( content.writeBody( body, bodySize ) != x509e_ok ) ) return x509e_internalError;

	return x509e_ok;
}

int p12_generateHMAC( cPTB &ptbHMAC, x509_OID &algOID, x509_struct &salt, x509_integer &sInt, cPassword *pwd, BYTE *message, int messageSize )
{
	if ( ( pwd == NULL ) || ( pwd->GenerateUnicodeString() == 0 ) )
		return ppe_invalidPassword;

	UINT iIterations;
	BYTE haID;
	
	int res = P12_getPBEAlgorithmInfo( algOID, salt, sInt, &iIterations, &haID );
	if ( res != x509e_ok ) return res;

	//
	//	SHA1 MAC-ing is implemented only.
	//
	if ( haID != pgphaSHA1 )
		return x509e_invalidPkt;

	int digestSize = 0;
	res = ha_getDigestSize( &digestSize, haID );
	if ( res != x509e_ok ) return res;

	cPTB ptb;

	int macType = P12_prgm4MAC;
	int macKeySize = digestSize; // Our assumption, there is no words in doc. PKCS #12: Personal Information Exchange Syntax v1.1

	if ( ptb.Allocate( macKeySize ) == NULL )
		return x509e_memoryAllocation;

	res = P12_producePseudoRandomBytes( ptb.buffer, macKeySize, salt.B, salt.L, iIterations, pwd->unicodeBuffer, pwd->unicodeBufferSize, (BYTE)macType, haID );
	if ( res != x509e_ok ) return res;

	if ( ptbHMAC.Allocate( digestSize ) == NULL )
		return x509e_memoryAllocation;

	res = hmac_sha1( message, messageSize, ptb.buffer, macKeySize, ptbHMAC.buffer );

	return res;
}

int sMacData::writeBody( BYTE *salt, int saltLen, cPassword *pwd, BYTE *message, int messageSize )
{
	int res = macSalt.writeBody( salt, saltLen ); 
	if ( res != x509e_ok ) return res;

	res = iterations.writeBody();
	if ( res != x509e_ok ) return res;

	cPTB ptbHMAC;
	res = p12_generateHMAC( ptbHMAC, mac.sAlgID.oid, macSalt, iterations, pwd, message, messageSize );
	if ( res != x509e_ok ) return res;

	res = mac.writeBody( ptbHMAC.buffer, ptbHMAC.bufferSize );
	if ( res != x509e_ok ) return res;

	return res;
}

int sMacData::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize, cPassword *pwd, BYTE *message, int messageSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	BYTE *pos = B;
	int size = L;

	res = mac.parse			( pos, size, &pos, &size ); if ( res != x509e_ok ) return res;
	res = macSalt.parse		( pos, size, &pos, &size ); if ( res != x509e_ok ) return res;
	res = iterations.parse	( pos, size, &pos, &size ); if ( res != x509e_ok ) return res;

	if ( size != 0 ) return x509e_invalidPktSize;

	cPTB ptbHMAC;
	res = p12_generateHMAC( ptbHMAC, mac.sAlgID.oid, macSalt, iterations, pwd, message, messageSize );
	if ( res != x509e_ok ) return res;

	if ( memcmp( mac.sValue.B, ptbHMAC.buffer, mac.sValue.L ) != 0 )
		return ppe_invalidPassword;

	return x509e_ok;
}

int p12_pfx_pdu_p7data::prepareForWritting( int contentInfoSize, const char *szOid, int saltLen, int iterations  )
{
	// OPTIONAL macData
	bNoMacData = ( szOid == NULL );
	int hashLen = 0;

	if ( !bNoMacData )
	{
		BYTE haID = (BYTE)-1;
		int res = P12_getAlgIDs( szOid, &haID ) ;
		if ( res != x509e_ok ) return res;

		res = ha_getDigestSize( &hashLen, haID );
		if ( res != x509e_ok ) return res;
	}

	return 
		x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			ver.prepareForWritting( 3 )
			+
			seqOfContInfo.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				p7dataOID.prepareForWritting( szP7data )
				+
				optional.prepareForWritting( asn1tag_CONSTRUCTED_OPTIONAL,
				octetStr.prepareForWritting( asn1tag_OCTET_STRING, contentInfoSize )
				)
			)
			+
			( bNoMacData ? 0 : macData.prepareForWritting( szOid, hashLen, saltLen, iterations  ) ) 
		);
}

int p12_pfx_pdu_p7data::writeHeader( BYTE *buff, int buffSize, int *pPos )
{
	if ( x509_struct::writeHeader( buff, buffSize, pPos ) == 0 ) return x509e_internalError;

	int pos = 0;
	ver.writeHeader( B, L, &pos );
	ver.writeBody();

	if ( seqOfContInfo.writeHeader( B, L, &pos ) == 0 )return x509e_internalError;
	else
	{
		int pos = 0;
		p7dataOID.writeHeader( seqOfContInfo.B, seqOfContInfo.L, &pos );
		p7dataOID.writeBody();

		if ( optional.writeHeader( seqOfContInfo.B, seqOfContInfo.L, &pos ) == 0 ) return x509e_internalError;
		if ( octetStr.writeHeader( optional.B, optional.L ) == 0 ) return x509e_internalError;
	}

	if ( !bNoMacData && ( macData.writeHeader( B, L, &pos ) == 0 ) ) return x509e_internalError;

	 return x509e_ok;
}

int p12_pfx_pdu_p7data::generateHMAC( const BYTE *salt, int saltLen, cPassword *pwd )
{
	//
	//	Octet body may be filled before by writeBody or direct writting.
	//
	if ( octetStr.B == NULL )
		return x509e_internalError;

	return macData.writeBody( (BYTE*)salt, saltLen, pwd, octetStr.B, octetStr.L );
}

int p12_pfx_pdu_p7data::writeBody( const BYTE *contentInfo, int ciSize )
{
	return octetStr.writeBody( (BYTE*)contentInfo, ciSize );
}

//	PKCS #7 
//	9.1 SignedData type
//
//	SignedData ::= SEQUENCE {
//	  version Version,
//	  digestAlgorithms DigestAlgorithmIdentifiers,
//	  contentInfo ContentInfo,
//	  certificates
//    	[0] IMPLICIT ExtendedCertificatesAndCertificates OPTIONAL,
//	  crls
//		[1] IMPLICIT CertificateRevocationLists OPTIONAL,
//	  signerInfos SignerInfos }
//	DigestAlgorithmIdentifiers ::=
//	  SET OF DigestAlgorithmIdentifier
//	SignerInfos ::= SET OF SignerInfo

//000F  30 80        3075: . . SEQUENCE
//0011  02 01           1: . . . INTEGER 1 ( 1 )
//0014  31 0B          11: . . . SET
//0016  30 09           9: . . . . SEQUENCE
//0018  06 05           5: . . . . . OID { secSIG-SHA1 }
//001F  05 00           0: . . . . . NULL
//0021  30 80          13: . . . SEQUENCE
//0023  06 09           9: . . . . OID { pkcs-7-data }
//002E  00 00           0: . . . . END OF CONTENTS MARKER	// here empty content
//0030  A0 82 08 D4  2260: . . . [0]
//0034  30 82 04 66  1126: . . . . SEQUENCE
//0038  30 82 02 4E   590: . . . . . SEQUENCE
//003C  A0 03           3: . . . . . . [0]
//003E  02 01           1: . . . . . . . INTEGER 2 ( 2 )

int p7_signedData::parse( const BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_CONSTRUCTED_SEQUENCE );
	if ( res != x509e_ok ) return res;

	BYTE *pos = B;
	int size = L;

	res = version.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( version.compare( 1 ) != 0 )
		return x509e_invalidPktVersion;

	res = digestAlgorithms.parse( pos, size, &pos, &size, asn1tag_CONSTRUCTED_SET );
	if ( res != x509e_ok ) return res;

	res = contentInfo.parse( pos, size, &pos, &size, asn1tag_CONSTRUCTED_SEQUENCE);
	if ( res != x509e_ok ) return res;

	x509_struct optional;
	res = optional.parse( pos, size );
	if ( res != x509e_ok ) return res;

	if ( optional.isOptional() && ( optional.getOptionalNum() == 0 ) )
	{
		res = certificates.parse( pos, size, &pos, &size );
		if ( res != x509e_ok ) return res;
	}

	res = optional.parse( pos, size );
	if ( res != x509e_ok ) return res;

	if ( optional.isOptional() && ( optional.getOptionalNum() == 1 ) )
	{
		res = crls.parse( pos, size, &pos, &size );
		if ( res != x509e_ok ) return res;
	}

	res = signerInfos.parse( pos, size, &pos, &size, asn1tag_CONSTRUCTED_SET );
	if ( res != x509e_ok ) return res;

	if ( size != 0 )
	{
		x509_struct eocm;
		res = eocm.parse( pos, size, &pos, &size, asn1tag_END_OF_CONTENTS_MARKER );
		if ( res != x509e_ok ) return res;
	}

	if ( size != 0 ) return x509e_invalidPktSize;

	return x509e_ok;
}

//	SafeBag ::= SEQUENCE {
//		bagId           BAG-TYPE.&id ({PKCS12BagSet}),
//		bagValue        [0] EXPLICIT BAG-TYPE.&Type({PKCS12BagSet}{@bagId}),
//		bagAttributes   SET OF PKCS12Attribute OPTIONAL
//	}
int p12_safeBag::parse( const BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize, const char **oidT )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_CONSTRUCTED_SEQUENCE );
	if ( res != x509e_ok )
		return res;

	BYTE *pos = B;
	int size = L;

	res = oid.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( ( oidT != NULL ) && ( !oid.check( oidT ) ) )
		return x509e_unexpectedOIDType;

	res = value.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( size == 0 ) // skip optional
		return x509e_ok;

	res = attr.parse( pos, size, &pos, &size );

	if ( size != 0 )
		return x509e_invalidPktSize;

	return res;
}

const char *szP7data = "1.2.840.113549.1.7.1";
const char *szP7encryptedData = "1.2.840.113549.1.7.6";

p12_safeBag_p7data::p12_safeBag_p7data( const char *bagType )
{
	if ( bagType == NULL )
		bagType = szP7data;

	m_bagType = bagType;

	//
	// The szP7data uses octet string inside optional value
	//	Others bag types write directly to optional value.
	//
	m_bUseOctet = strcmp( szP7data, m_bagType ) == 0;
}

int p12_safeBag_p7data::prepareForWritting( int contentInfoSize, const p12_attribute *attrList )
{
	m_attrList = attrList;
	
	return 
		x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			oid.prepareForWritting( m_bagType )
			+
			value.prepareForWritting( asn1tag_CONSTRUCTED_OPTIONAL,
				m_bUseOctet ?
					content.prepareForWritting( asn1tag_OCTET_STRING, contentInfoSize ) :
					contentInfoSize
			)
			+ ( attrList == 0 ? 0 :
				attr.prepareForWritting( asn1tag_CONSTRUCTED_SET, p12_attribute::getSize( attrList ) )
			 )
		);
}

int p12_safeBag_p7data::writeHeader( BYTE *buff, int buffSize, int *pPos )
{
	int hdrSize = x509_struct::writeHeader( buff, buffSize, pPos );
	if ( hdrSize == 0 ) return 0;

	int pos = 0;
	if ( oid.writeHeader( B, L, &pos ) == 0 ) return 0;
	oid.writeBody();

	if ( value.writeHeader( B, L, &pos ) == 0 ) return 0;

	if ( m_bUseOctet && ( content.writeHeader( value.B, value.L ) == 0 ) )
		return 0;

	if ( m_attrList != 0 )
	{
		if ( attr.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		int res = p12_attribute::writeBodies( attr, m_attrList );
		if ( res != x509e_ok ) return 0;
	}

	if ( pos != (int)L )
		return 0;

	return hdrSize;
}

int p12_safeBag_p7data::writeBody( const BYTE *contentInfo, int ciSize )
{
	int res = 0;

	if ( m_bUseOctet )
		res = content.writeBody( contentInfo, ciSize );
	else
		res = value.writeBody( contentInfo, ciSize );

	return res;
}


//static
int p12_attribute::getSize( const p12_attribute *attrList )
{
	int attrSize = 0;

	for( const p12_attribute *attrUnit = attrList; attrUnit; attrUnit = attrUnit->next )
	{
		x509_struct seq;
		x509_struct  oid;
		x509_struct  set;
		x509_struct   val;

		attrSize += seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			oid.prepareForWritting( asn1tag_OBJECT_IDENTIFIER, string2oid( NULL, 0, attrUnit->oidStr ) )
			+
			set.prepareForWritting( asn1tag_CONSTRUCTED_SET, 
				val.prepareForWritting( attrUnit->type, attrUnit->size )
			)
		);
	}

	return attrSize;
}

//static
int p12_attribute::writeBodies( x509_struct &attr, const p12_attribute *attrList )
{
	int pos = 0;

	for( const p12_attribute *attrUnit = attrList; attrUnit; attrUnit = attrUnit->next )
	{
		x509_struct seq;
		x509_OID	oid;
		x509_struct  set;
		x509_struct   val;

		// v. 1.04.8.1 Bug #2856 Key Manager crashes
		int attrSize = seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			oid.prepareForWritting( attrUnit->oidStr )
			+
			set.prepareForWritting( asn1tag_CONSTRUCTED_SET, 
				val.prepareForWritting( attrUnit->type, attrUnit->size )
			)
		);

		if ( attrSize == -1 ) return x509e_internalError;

		int lPos = 0;

		if ( seq.writeHeader( attr.B, attr.L, &pos ) == 0 )				return x509e_internalError;
		 if ( oid.writeHeader( seq.B, seq.L, &lPos ) == 0 )				return x509e_internalError;
		  if ( oid.writeBody() != ppe_ok )								return x509e_internalError;
		 if ( set.writeHeader( seq.B, seq.L, &lPos ) == 0 )				return x509e_internalError;
		  if ( val.writeHeader( set.B, set.L ) == 0 )					return x509e_internalError;
		   if ( val.writeBody( attrUnit->value, attrUnit->size ) != ppe_ok ) return x509e_internalError;

		if ( lPos != (int)seq.L ) return x509e_internalError;
	}

	return x509e_ok;
}
