/////////////////////////////////////////////////////////////////
//
//	Copyright (c) 2015 Jetico, Inc., Finland
//	All rights reserved.
//
//	File: pkcs12_structs.h
//	Revision: 
//	Created:
//	Description:
//
/////////////////////////////////////////////////////////////////

#pragma once

#include "x509_structs.h"
#include "../SMemory/cPassword.h"

struct sPbeParam : public x509_struct
{
	// asn1tag_CONSTRUCTED_SEQUENCE
	x509_OID		oid;
	x509_struct		seqParams;
	x509_struct			octSalt; // asn1tag_OCTET_STRING
	x509_integer		intIterations;

	int prepareForWritting( const char *pbeAlgOID, unsigned int saltSize, unsigned int iterations );
	int writeHeader( BYTE *buff, int buffSize, int *pPos  );
	int writeBody( const BYTE *salt, unsigned int saltSize );
};
	
struct sEncryptedContentInfo : public x509_struct
{
	// asn1tag_CONSTRUCTED_SEQUENCE

	x509_integer	version;	// 0 for this version of the document.
	x509_struct		seq;		// asn1tag_CONSTRUCTED_SEQUENCE
	x509_OID			oid;	// szP7data
	sPbeParam			pbe;
	x509_struct			content;// Optional; Encrypted data content has tag asn1tag_OPTIONAL not asn1tag_CONSTRUCTED

	int prepareForWritting( const char *pbeAlgOID, unsigned int saltSize, unsigned int iterations, unsigned int bodySize );
	int writeHeader( BYTE *buff, int buffSize, int *pPos = NULL );
	int writeBody( const BYTE *salt, unsigned int saltSize, const BYTE *body, unsigned int bodySize );
};


//
// MacData ::= SEQUENCE {
//    mac         DigestInfo,
//    macSalt     OCTET STRING,
//    iterations  INTEGER DEFAULT 1
//    -- Note: The default is for historical reasons and its
//    --       use is deprecated.
//

struct sMacData : public x509_struct
{
	x509_digestInfo	mac;
	x509_struct		macSalt;
	x509_integer	iterations;

	int prepareForWritting( const char *szOid, int hashLen, int saltLen, int iterationsQ )
	{
		return x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			mac.prepareForWritting( szOid, hashLen ) +
			macSalt.prepareForWritting( asn1tag_OCTET_STRING, saltLen ) +
			iterations.prepareForWritting( iterationsQ ) );
	}

	int writeHeader( BYTE *buff, int buffSize, int *pPos = NULL )
	{
		int hdrSize = x509_struct::writeHeader( buff, buffSize, pPos );
		if ( hdrSize == 0 ) return 0;

		int pos = 0;
		if ( !mac.writeHeader		( B, L, &pos ) ) return 0;
		if ( !macSalt.writeHeader	( B, L, &pos ) ) return 0;
		if ( !iterations.writeHeader( B, L, &pos ) ) return 0;

		return hdrSize;
	}

	int writeBody( BYTE *salt, int saltLen, cPassword *pwd, BYTE *message, int messageSize );

	int parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize, cPassword *pwd, BYTE *message, int messageSize );
};

struct p12_pfx_pdu_p7data : public x509_struct
{
	x509_integer	ver;
	x509_struct		seqOfContInfo;
	x509_OID		p7dataOID;
	x509_struct		optional;
	x509_struct		octetStr;
	sMacData		macData; // OPTIONAL

	bool bNoMacData;

	int prepareForWritting( int contentInfoSize, const char *szOid = NULL, int saltLen = 0, int iterations = 0 );
	int writeHeader( BYTE *buff, int buffSize, int *pPos = NULL );
	int writeBody( const BYTE *contentInfo, int contentInfoSize );

	//
	//	generateHMAC
	//		Generates SHA1 HMAC sequence using octetString filled data.
	//
	int generateHMAC( const BYTE *salt, int saltLen, cPassword *pwd );
};

struct p12_attribute
{
	const char *oidStr;
	BYTE type;
	BYTE *value;
	int  size;

	struct p12_attribute *next;

	p12_attribute( const char *pOidStr, BYTE pType, BYTE *pValue, int pSize, p12_attribute *pNext = NULL )
	{
		oidStr	= pOidStr;
		type	= pType;
		value	= pValue;
		size	= pSize;
		next	= pNext;
	}


	p12_attribute() { memset( this, 0, sizeof( *this ) ); }
	~p12_attribute() { if ( next ) { delete next; next = NULL; } }
	static void freeList( p12_attribute *list ) { if ( list ) delete list; }

	static int getSize( const p12_attribute *attrList );

	//
	//	writeBodies
	//		writes headers and bodies for all attributes.
	//
	static int writeBodies( x509_struct &attr, const p12_attribute *attrList );
};


//	SafeBag ::= SEQUENCE {
//		bagId           BAG-TYPE.&id ({PKCS12BagSet}),
//		bagValue        [0] EXPLICIT BAG-TYPE.&Type({PKCS12BagSet}{@bagId}),
//		bagAttributes   SET OF PKCS12Attribute OPTIONAL
//	}
struct p12_safeBag : public x509_struct // SEQUENCE
{
	x509_OID	oid;	// OID
	x509_struct value;	// EXPLICIT 
	x509_struct attr;	// SET OPTIONAL
	const p12_attribute *m_attrList;

	p12_safeBag()
	{
		m_attrList = NULL;
	}

	//
	// parses packet and fills values
	//
	//	oidT sample:
	//	char *oidT[] = { "pkcs", "7", "signedData", NULL };
	//	if is not null the func checks the oid
	//	by  oid.check( oidT ) if false it returns x509e_unexpectedOIDType
	//
	int parse( const BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL, const char **oidT = NULL );
};

//
//	p12_safeBag_p7data
//
extern const char *szP7data;
extern const char *szP7encryptedData;

struct p12_safeBag_p7data : public p12_safeBag
{
	const char *m_bagType;
	bool m_bUseOctet;

	x509_struct	content; // octetStr for szP7data, sequence for szP7encryptedData;

	p12_safeBag_p7data( const char *bagType = NULL );

	int prepareForWritting( int contentInfoSize, const p12_attribute *attrList = NULL );
	int writeHeader( BYTE *buff, int buffSize, int *pPos = NULL );
	int writeBody( const BYTE *contentInfo, int ciSize );

	int getContentSize()
	{	return m_bUseOctet ? content.L : value.L; 	}

	BYTE *getContentBody()
	{	return m_bUseOctet ? content.B : value.B; 	}
};


//	PKCS #7 
//	9.1 SignedData type
//
//	SignedData ::= SEQUENCE {
//	  version Version,
//	  digestAlgorithms DigestAlgorithmIdentifiers,
//	  contentInfo ContentInfo,
//	  certificates
//    	[0] IMPLICIT ExtendedCertificatesAndCertificates OPTIONAL,
//	  crls
//		[1] IMPLICIT CertificateRevocationLists OPTIONAL,
//	  signerInfos SignerInfos }
//	DigestAlgorithmIdentifiers ::=
//	  SET OF DigestAlgorithmIdentifier
//	SignerInfos ::= SET OF SignerInfo

struct p7_signedData : public x509_struct
{
	x509_integer	version;
	x509_struct		digestAlgorithms;	// SET
	x509_struct		contentInfo;		// SEQUENCE
	x509_struct		certificates;		// [0] optional
	x509_struct		crls;				// [1] optional
	x509_struct		signerInfos;		// SET

	int parse( const BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );
};

//
// P12_makeSafeContents
//	every safeBagsT unit has to be prepared before, f.e by P12_makePKCS8ShroudedKeyBagPkt
//
int P12_makeSafeContents( BYTE *SC, int *pSCSize, cPlainTextBlock *safeBagsT, int safeBagsTSize );

struct p12_safeContent : public x509_struct
{
	//
	//	SafeContents ::= SEQUENCE OF SafeBag
	//
	int prepareForWritting( cPlainTextBlock *safeBagsT, int safeBagsTSize );
	int prepareForWritting( x509_struct *safeBagsT, int safeBagsTSize );
	int writeBody( cPlainTextBlock *safeBagsT, int safeBagsTSize );
};

class cAttrAutoFree
{
public:
	
	p12_attribute *head;

	cAttrAutoFree( p12_attribute *list = NULL )
	{
		head = NULL;

		if ( list != NULL )
			set( list );
	}

	void set( p12_attribute *list )
	{
		if ( head != NULL )
			p12_attribute::freeList( head );

		head = list;
	}

	~cAttrAutoFree()
	{
		if ( head != NULL )
			p12_attribute::freeList( head );

		head = NULL;
	}

};